#!/usr/bin/env python2
import pygame, argparse, time, math, json, os

formatversion = 2

CLICKDIST = 10
DRAGDIST = 5

ap = argparse.ArgumentParser()
ap.add_argument('image')
args = ap.parse_args()

image = pygame.image.load(args.image)
screen = pygame.display.set_mode((1024, 768))
pygame.display.set_caption('lowpoly editor')

camera = (image.get_width() / 2., image.get_height() / 2.)
zoom = screen.get_width() / float(image.get_width())

sizedimage = None
sizedzoom = None
sizedpos = None
def drawbg():
    global sizedimage, sizedzoom, sizedpos
    if zoom != sizedzoom:
        if zoom <= 1:
            size = intpos((image.get_width() * zoom, image.get_height() * zoom))
            sizedimage = pygame.transform.smoothscale(image, size)
            sizedzoom = zoom
            sizedpos = (0, 0)
        else:
            a = screen2image((0, 0))
            b = screen2image(screen.get_size())
            r = pygame.Rect((a[0], a[1], b[0] - a[0], b[1] - a[1]))
            rc = image.get_rect().clip(r)
            size = intpos((rc.w * zoom, rc.h * zoom))
            sizedimage = pygame.transform.smoothscale(image.subsurface(rc), size)
            sizedzoom = (zoom, camera)
            sizedpos = (rc.x, rc.y)

    screen.blit(sizedimage, intpos(image2screen(sizedpos)))

def screen2image(pos):
    w, h = screen.get_size()
    return ((pos[0] - w/2) / zoom + camera[0], (pos[1] - h/2) / zoom + camera[1])

def image2screen(pos):
    w, h = screen.get_size()
    return ((pos[0] - camera[0]) * zoom + w/2, (pos[1] - camera[1]) * zoom + h/2)

def scale2image(n):
    return n / zoom

def intpos(pos):
    return (int(pos[0]), int(pos[1]))

class Point:
    def __init__(self, pos):
        self.x = pos[0]
        self.y = pos[1]
        
    @property
    def pos(self):
        return intpos(image2screen((self.x, self.y)))

class Poly:
    def __init__(self):
        self.pts = []
        self.color = (255, 255, 255)

    @property
    def poslist(self):
        return [pt.pos for pt in self.pts]

    @property
    def poslistraw(self):
        return [intpos((pt.x, pt.y)) for pt in self.pts]

def save(filename):
    delbackup = False
    if os.path.exists(filename):
        os.rename(filename, filename + 'backup')
        delbackup = True
        
    f = open(filename, 'w')
    d = {
        'version': formatversion,
        'points': [(pt.x, pt.y) for pt in points],
        'polys': [{'color': p.color, 'pts': [points.index(pt) for pt in p.pts]} for p in polys]
    }
    json.dump(d, f)
    f.close()
    
    if delbackup:
        os.remove(filename + 'backup')

def load(filename):
    f = open(filename, 'r')
    d = json.load(f)
    f.close()

    fileversion = d.get('version', 1)
    if fileversion != formatversion:
        print 'FATAL: %s has format version %d (expecting %d)' % (filename, fileversion, formatversion)
        exit(1)

    for pos in d['points']:
        points.append(Point(pos))
    for poly in d['polys']:
        p = Poly()
        p.color = tuple(poly['color'])
        p.pts = [points[i] for i in poly['pts']]
        polys.append(p)

def render():
    screen.fill((255, 0, 255))
    if not hidepic:
        drawbg()
    
    for poly in polys:
        pygame.draw.polygon(screen, poly.color, poly.poslist, 2 if wireframe else 0)

    if showpoints:
        for pt in points:
            if pt in newpoly.pts:
                if len(newpoly.pts) == 1:
                    color = (0, 0, 255)
                else:
                    color = (255, 0, 0)
            else:
                color = (0, 255, 0)
            pygame.draw.circle(screen, color, pt.pos, 3)
        
    if selpoly != None:
        pygame.draw.polygon(screen, (0, 0, 255), selpoly.poslist, 2)
        
        pygame.draw.line(screen, (255, 0, 0), (10, 10), (10, 110), 5)
        c = 10 + 100 - selpoly.color[0] * 100 / 255
        pygame.draw.line(screen, (0, 0, 0), (10, c), (15, c), 2)
        
        pygame.draw.line(screen, (0, 255, 0), (30, 10), (30, 110), 5)
        c = 10 + 100 - selpoly.color[1] * 100 / 255
        pygame.draw.line(screen, (0, 0, 0), (30, c), (35, c), 2)
        
        pygame.draw.line(screen, (0, 0, 255), (50, 10), (50, 110), 5)
        c = 10 + 100 - selpoly.color[2] * 100 / 255
        pygame.draw.line(screen, (0, 0, 0), (50, c), (55, c), 2)
        
    pygame.display.update()

def saveimage(filename):
    global wireframe
    wireframe = False
    surf = pygame.Surface(image.get_size(), pygame.SRCALPHA)
    surf.fill((0, 0, 0, 0))

    for poly in polys:
        pygame.draw.polygon(surf, poly.color, poly.poslistraw)
        
    pygame.image.save(surf, filename)

def getcolorat(pos):
    try:
        return image.get_at(intpos(pos))[:3]
    except IndexError:
        return (255, 0, 255)

def getnearest(pos, maxdist=None):
    best = None
    bestpt = None
    for pt in points:
        d = math.hypot(pt.x - pos[0], pt.y - pos[1])
        if (maxdist == None or d <= maxdist) and (best == None or d < best):
            best = d
            bestpt = pt
    return bestpt

def getpolycenter(poly):
    return (sum(pt.x for pt in poly.pts) / 3, sum(pt.y for pt in poly.pts) / 3)

def getnearestpoly(pos):
    best = None
    bestpoly = None
    for poly in polys:
        center = getpolycenter(poly)
        d = math.hypot(center[0] - pos[0], center[1] - pos[1])
        if best == None or d < best:
            best = d
            bestpoly = poly
    return bestpoly

def colorwidgetpos(pos):
    '''
    return (x, y) or None
    x will be in [0, 1, 2]
    y will be between 0 and 1
    '''
    x, y = image2screen(pos)
    if x < 60 and y < 120:
        x = int(x / 20)
        y = 1 - (y - 10) / 100.
        if y < 0:
            y = 0
        elif y > 1:
            y = 1
        return (x, y)
    return None

def dragging(button, pos, rel):
    global movepoint, dirty, camera

    if button == 1:
        # dragging in color widget
        if selpoly != None and colorwidgetpos(mousedown[0]):
            if colorwidgetpos(pos):
                click(pos)
            return

        if movepoint == None:
            movepoint = getnearest(pos)
        movepoint.x += rel[0]
        movepoint.y += rel[1]
        dirty = True

    elif button == 2:
        camera = (camera[0] - rel[0], camera[1] - rel[1])
        dirty = True

def click(pos):
    global points, newpoly, dirty, selpoly

    if selpoly != None and colorwidgetpos(pos):
        i, cf = colorwidgetpos(pos)
        c = list(selpoly.color)
        c[i] = 256 * cf
        if c[i] < 0:
            c[i] = 0
        elif c[i] > 255:
            c[i] = 255
        selpoly.color = tuple(c)
        dirty = True
        return
    
    if len(points) > 0:
        selpoly = None
        pt = getnearest(pos, scale2image(CLICKDIST))
        if pt != None and pt not in newpoly.pts:
            newpoly.pts.append(pt)
            if len(newpoly.pts) == 3:
                newpoly.color = getcolorat(getpolycenter(newpoly))
                selpoly = newpoly
                polys.append(newpoly)
                newpoly = Poly()
            dirty = True
            return
    
    points.append(Point(pos))
    dirty = True

def scrollwheel(dir):
    global zoom, dirty
    if dir > 0:
        zoom *= 1.1
    else:
        zoom /= 1.1
    dirty = True

def rightclick(pos):
    global selpoly, dirty
    if len(polys) > 0:
        selpoly = getnearestpoly(pos)
        dirty = True

def keypress(key):
    global newpoly, dirty, selpoly, wireframe, hidepic, showpoints
    if key == pygame.K_o:
        if selpoly != None:
            pos = screen2image(pygame.mouse.get_pos())
            selpoly.color = getcolorat(pos)
            dirty = True
    elif key == pygame.K_DELETE:
        if selpoly != None:
            polys.remove(selpoly)
            selpoly = None
            dirty = True
        elif len(newpoly.pts) > 0:
            for pt in newpoly.pts:
                points.remove(pt)
                for poly in polys[:]:
                    if pt in poly.pts:
                        polys.remove(poly)
            newpoly = Poly()
            dirty = True
    elif key == pygame.K_w:
        wireframe = not wireframe
        dirty = True
    elif key == pygame.K_h:
        hidepic = not hidepic
        dirty = True
    elif key == pygame.K_p:
        showpoints = not showpoints
        dirty = True
    elif key == pygame.K_s:
        save(filename)
        print 'wrote', filename
    elif key == pygame.K_MINUS:
        if selpoly != None:
            polys.remove(selpoly)
            polys.insert(0, selpoly)
            dirty = True
    elif key == pygame.K_EQUALS:
        if selpoly != None:
            polys.remove(selpoly)
            polys.append(selpoly)
            dirty = True
    elif key == pygame.K_SPACE:
        selpoly = None
        newpoly = Poly()
        dirty = True

dirty = True
running = True

points = []
polys = []
newpoly = Poly()
selpoly = None

wireframe = False
hidepic = False
showpoints = True

mousedown = None
mousedrag = False
movepoint = None

fileroot = os.path.splitext(args.image)[0]
filename = fileroot + '.poly'
if os.path.exists(filename):
    load(filename)

while running:
    for event in pygame.event.get():
        if event.type == pygame.MOUSEBUTTONDOWN:
            pos = screen2image(event.pos)
            if event.button in (1, 2):
                mousedown = (pos, event.button)
            elif event.button == 3:
                rightclick(pos)
            elif event.button in (4, 5):
                scrollwheel(-1 if event.button == 5 else 1)
        elif event.type == pygame.MOUSEMOTION:
            pos = screen2image(event.pos)
            rel = (scale2image(event.rel[0]), scale2image(event.rel[1]))
            if mousedown and not mousedrag and math.hypot(pos[0] - mousedown[0][0], pos[1] - mousedown[0][1]) > scale2image(DRAGDIST):
                mousedrag = True
            if mousedrag:
                dragging(mousedown[1], pos, rel)
        elif event.type == pygame.MOUSEBUTTONUP:
            if event.button == 1:
                if not mousedrag:
                    click(mousedown[0])
            mousedown = None
            mousedrag = False
            movepoint = None
        elif event.type == pygame.KEYDOWN:
            keypress(event.key)
        elif event.type == pygame.QUIT:
            running = False
    if dirty:
        render()
        dirty = False
    time.sleep(.01)

save(filename)
saveimage(fileroot + '_poly.png')
